/*
  Test.h - Test library for Wiring - implementation
  Copyright (c) 2006 John Doe.  All right reserved.
*/


// include this library's description file
#include "PropWiFi.h"


// Constructor /////////////////////////////////////////////////////////////////
// Function that handles the creation and setup of instances




void PropWiFi::enableHeartbeating(bool state)
{
	mEnableHeartbeating = state;
}


PropWiFi::PropWiFi(Stream *serial, char *ssid, char *pwd, byte id, int udpPort)
{
	mStatus = WL_IDLE_STATUS;
	WiFi.init(serial);
	mSsidBuf[31] = '\0';
	for (int i = 0; i < 31; i++)
	{
		mSsidBuf[i] = ssid[i];
		if (ssid[i] == '\0')
			break;
	}
	for (int i = 0; i < 31; i++)
	{
		mPwdBuf[i] = pwd[i];
		if (pwd[i] == '\0')
			break;
	}
	mUdpPort = udpPort;
	mId = id;
	mLastConnCheck = 0;
	for(int i = 0; i < 8; i++)
	{
		mMsgQueue[i].idx = 0;
		mHeartbeatTracking[i] = 0;
	}
	mConnAchievedCB = NULL;
	mConnLostCB = NULL;
	mMsgReceivedCB = NULL;
	mMsgAckdCB = NULL;
	mHeartbeatDiscoveredCB = NULL;
	mHeartbeatLostCB = NULL;
	mResendTimer = 0;
	mEnableHeartbeating = false;
	mHeartbeatTimer = 0;
	mHeartbeatListening[0] = 0;
	mHeartbeatListening[1] = 0;
	mHeartbeatListening[2] = 0;
	mHeartbeatListening[3] = 0;
	mHeartbeatListening[4] = 0;
	mHeartbeatListening[5] = 0;
	mHeartbeatListening[6] = 0;
	mHeartbeatListening[7] = 0;
	mMsgIdx = 1;
	mBCast = IPAddress(255,255,255,255);
	beginConnection();
}

void PropWiFi::sendHeartbeat()
{
	byte actualMsg[4];
	actualMsg[0] = mId;
	actualMsg[1] = 255;
	actualMsg[2] = 1;
	actualMsg[3] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);
	mUdp.write(actualMsg, 4);
	mUdp.endPacket();	
}

void PropWiFi::sendMessageAck(byte recipient, byte msgIdx)
{
	byte actualMsg[5];
	actualMsg[0] = mId;
	actualMsg[1] = recipient;
	actualMsg[2] = 3;
	actualMsg[3] = msgIdx;
	actualMsg[4] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);
	mUdp.write(actualMsg, 5);
	mUdp.endPacket();
}

void PropWiFi::sendMessageHelper(byte recipient, byte msgIdx, unsigned long data)
{
	byte actualMsg[9];
	actualMsg[0] = mId;
	actualMsg[1] = recipient;
	actualMsg[2] = 2;
	actualMsg[3] = msgIdx;
	actualMsg[4] = ((byte *)&data)[0];
	actualMsg[5] = ((byte *)&data)[1];
	actualMsg[6] = ((byte *)&data)[2];
	actualMsg[7] = ((byte *)&data)[3];
	actualMsg[8] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);
	mUdp.write(actualMsg, 9);
	mUdp.endPacket();
}

void PropWiFi::resendMessages()
{
	unsigned long curMillis = millis();
	for (int i = 0; i < 8; i++)
	{
		if (mMsgQueue[i].idx != 0)
		{
			if (((curMillis < mMsgQueue[i].resent) && (curMillis > propWiFiResendTime)) ||
				(curMillis > mMsgQueue[i].resent + propWiFiResendTime))
			{
				mMsgQueue[i].resent = curMillis;
				sendMessageHelper(mMsgQueue[i].recipient, mMsgQueue[i].idx, mMsgQueue[i].payload);
				if (((curMillis < mMsgQueue[i].sent) && (curMillis > propWiFiMaxResendTime)) ||
					(curMillis > mMsgQueue[i].sent + propWiFiMaxResendTime))
				{
					mMsgQueue[i].idx = 0;
				}
			}
		}
	}
}

void PropWiFi::registerMessageAckdCB(void (*msgAckdCB)(byte msgId))
{
	mMsgAckdCB = msgAckdCB;
}

void PropWiFi::registerMessageReceivedCB(void (*msgReceivedCB)(byte source, unsigned long data))
{
	mMsgReceivedCB = msgReceivedCB;
}

byte PropWiFi::sendMessage(byte recipient, unsigned long data)
{
	int queueIdx = 0;
	for (queueIdx = 0; queueIdx < 8; queueIdx++)
	{
		if (mMsgQueue[queueIdx].idx == 0)
		{
			mMsgQueue[queueIdx].idx = mMsgIdx;
			mMsgQueue[queueIdx].payload = data;
			mMsgQueue[queueIdx].sent = millis();
			mMsgQueue[queueIdx].resent = mMsgQueue[queueIdx].sent;
			mMsgQueue[queueIdx].recipient = recipient;
			break;
		}
	}
	//Actually send message...
	sendMessageHelper(recipient, mMsgIdx, data);
	byte returnVal = mMsgIdx;
	mMsgIdx++;
	if (mMsgIdx == 0)
	{
		mMsgIdx = 1;
	}
	return returnVal;
}

void PropWiFi::beginConnection()
{
	if (mStatus != WL_CONNECTED)
	{
		mStatus = WiFi.begin(mSsidBuf, mPwdBuf);
		if (mStatus == WL_CONNECTED)
		{
			IPAddress myIp = WiFi.localIP();
			IPAddress subnetMask = WiFi.subnetMask();
			mBCast[0] = (subnetMask[0] & myIp[0]) | (~subnetMask[0]);
			mBCast[1] = (subnetMask[1] & myIp[1]) | (~subnetMask[1]);
			mBCast[2] = (subnetMask[2] & myIp[2]) | (~subnetMask[2]);
			mBCast[3] = (subnetMask[3] & myIp[3]) | (~subnetMask[3]);
			/*Serial.print(myIp[0]);
			Serial.print(".");
			Serial.print(myIp[1]);
			Serial.print(".");
			Serial.print(myIp[2]);
			Serial.print(".");
			Serial.println(myIp[3]);
			Serial.print(subnetMask[0]);
			Serial.print(".");
			Serial.print(subnetMask[1]);
			Serial.print(".");
			Serial.print(subnetMask[2]);
			Serial.print(".");
			Serial.println(subnetMask[3]);
			Serial.print(mBCast[0]);
			Serial.print(".");
			Serial.print(mBCast[1]);
			Serial.print(".");
			Serial.print(mBCast[2]);
			Serial.print(".");
			Serial.println(mBCast[3]);*/
			
			mUdp.begin(mUdpPort);
			if (mConnAchievedCB != NULL)
			{
				mConnAchievedCB();
			}
		}
		else
		{
			//Serial.println("Failure to connect!");
		}
	}
}

bool PropWiFi::checkConnection()
{
	int newStatus = WiFi.status();
	if (mStatus != newStatus)
	{
		if ((mStatus == WL_CONNECTED) && (mConnLostCB != NULL))
		{
			mConnLostCB();
		}
		else if (newStatus == WL_CONNECTED)
		{
			mUdp.begin(mUdpPort);
			if (mConnAchievedCB != NULL)
			{
				mConnAchievedCB();
			}
		}
		mStatus = newStatus;
	}
	if (mStatus != WL_CONNECTED)
	{
		beginConnection();
	}
}

void PropWiFi::registerConnectionAchievedCB(void (*connAchievedCB)())
{
	mConnAchievedCB = connAchievedCB;
}
void PropWiFi::registerConnectionLostCB(void (*connLostCB)())
{
	mConnLostCB = connLostCB;
}



void PropWiFi::processIncoming()
{
	int length = mUdp.parsePacket();
	if ((length > 9) || (length < 4))
	{		
		mUdp.flush();
	}
	else
	{
		byte packetBuff[11];
		packetBuff[9] = 0;
		int actualLength =  mUdp.read(packetBuff, length);
		if ((actualLength == 4) && (packetBuff[2] == 1))
		{
			//Handle heartbeat.
			unsigned long curMillis = 0;
			for (int i = 0; i < 8; i++)
			{
				if (mHeartbeatListening[i] == packetBuff[0])
				{
					if (curMillis == 0)
					{
						curMillis = millis();
						if (curMillis == 0)
							curMillis++;
					}
					if ((mHeartbeatDiscoveredCB != 0) && (mHeartbeatTracking[i] == 0))
					{
						mHeartbeatDiscoveredCB(mHeartbeatListening[i]);
					}
					mHeartbeatTracking[i] = curMillis;
				}
			}
		}
		else if ((actualLength == 5) && (packetBuff[2] == 3))
		{
			//Handle message ack.
			if (packetBuff[1] == mId)
			{
				//This was ack for a message from us...
				for(int i=0; i <  8; i++)
				{
					if (mMsgQueue[i].idx == packetBuff[3])
					{
						mMsgQueue[i].idx = 0;
					}
				}
				if (mMsgAckdCB != NULL)
				{
					mMsgAckdCB(packetBuff[3]);
				}
			}
		}
		else if ((actualLength == 9) && (packetBuff[2] == 2))
		{
			//Handle message.
			//Test if we are destination...
			if (packetBuff[1] == mId)
			{
				//This is a message for us...
				if (mMsgReceivedCB != NULL)
				{
					unsigned long dataBlock;
					((byte *)&dataBlock)[0] = packetBuff[4];
					((byte *)&dataBlock)[1] = packetBuff[5];
					((byte *)&dataBlock)[2] = packetBuff[6];
					((byte *)&dataBlock)[3] = packetBuff[7];
					mMsgReceivedCB(packetBuff[0], dataBlock);
				}
				sendMessageAck(packetBuff[0], packetBuff[3]);
			}
		}
		
	}
}
void PropWiFi::registerHeartbeatDiscoveredCB(void (*heartbeatDiscoveredCB)(byte source))
{
	mHeartbeatDiscoveredCB = heartbeatDiscoveredCB;
}

void PropWiFi::registerHeartbeatLostCB(void (*heartbeatLostCB)(byte source))
{
	mHeartbeatLostCB = heartbeatLostCB;
}

void PropWiFi::tick()
{
	unsigned long curMillis = millis();
	if (mStatus == WL_CONNECTED)
	{
		processIncoming();
	}
	if ((mLastConnCheck == 0) ||
		(curMillis > mLastConnCheck + propWiFiConnCheckTime) ||
		((curMillis < mLastConnCheck) && (curMillis > propWiFiConnCheckTime)))
	{
		mLastConnCheck = curMillis;
		checkConnection();
	}
	if ((mStatus == WL_CONNECTED) && ((curMillis > mResendTimer + propWiFiResendTime) ||
		((curMillis < mResendTimer) && (curMillis > propWiFiResendTime))))
	{
		mResendTimer = curMillis;
		resendMessages();
	}
	bool heartbeatTimerUp = false;
	if ((curMillis > mHeartbeatTimer + propWiFiHeartbeatTimer) ||
		((curMillis < mHeartbeatTimer) && (curMillis > propWiFiHeartbeatTimer)))
	{
		mHeartbeatTimer = curMillis;
		if ((mStatus == WL_CONNECTED) && mEnableHeartbeating)
		{
			sendHeartbeat();
		}
		for (int i = 0; i < 8; i++)
		{
			if ((mHeartbeatListening[i] != 0) && (mHeartbeatTracking[i] != 0))
			{
				if (((curMillis < mHeartbeatTracking[i]) && (curMillis > propWiFiHeartbeatTimeout)) ||
					(curMillis > mHeartbeatTracking[i] + propWiFiHeartbeatTimeout))
				{
					mHeartbeatTracking[i] = 0;
					if (mHeartbeatLostCB != NULL)
					{
						mHeartbeatLostCB(mHeartbeatListening[i]);
					}						
				}
			}
		}
	}
}

void PropWiFi::listenHeartbeat(byte source)
{
	for (int i = 0; i < 8; i++)
	{
		if (mHeartbeatListening[i] == 0)
		{
			mHeartbeatListening[i] = source;
			mHeartbeatTracking[i] = 0;
			return;
		}
	}
}
int PropWiFi::connectionStatus()
{
	return mStatus;
}