/*
  Test.h - Test library for Wiring - description
  Copyright (c) 2006 John Doe.  All right reserved.
*/

// ensure this library description is only included once
#ifndef PropWiFi_h
#define PropWiFi_h

#include <WiFiEsp.h>
#include <WiFiEspUdp.h>

#if ARDUINO >= 100
 #include "Arduino.h"
#else
 #include "WProgram.h"
#endif

#define propWiFiConnCheckTime 5000
#define propWiFiResendTime 200
#define propWiFiMaxResendTime 10000
#define propWiFiHeartbeatTimer 600
#define propWiFiHeartbeatTimeout 2500

// library interface description
class PropWiFi
{
  // user-accessible "public" interface
  public:
    PropWiFi(Stream *serial, char *ssid, char *pwd, byte id, int udpPort);
    void tick();

	void registerConnectionAchievedCB(void (*connAchievedCB)());
	void registerConnectionLostCB(void (*connLostCB)());
	void registerMessageReceivedCB(void (*msgReceivedCB)(byte source, unsigned long data));
	void registerMessageAckdCB(void (*msgAckdCB)(byte msgId));
	void registerHeartbeatDiscoveredCB(void (*heartbeatDiscoveredCB)(byte source));
	void registerHeartbeatLostCB(void (*heartbeatLostCB)(byte source));
	byte sendMessage(byte recipient, unsigned long data);
	
	void enableHeartbeating(bool state = true);
	
	void listenHeartbeat(byte source);
	
	int connectionStatus();
	
	
  // library-accessible "private" interface
  private:
	
	struct PropWiFiMsgQueueStruct {
		unsigned long sent;
		unsigned long resent;
		byte recipient;
		byte idx;
		unsigned long payload;
	};
  
	PropWiFiMsgQueueStruct mMsgQueue[8];
	char mSsidBuf[32];
	char mPwdBuf[32];
	byte mId;
	int mUdpPort;
	byte mMsgIdx;
	bool mEnableHeartbeating;
	byte mHeartbeatListening[8];
	WiFiEspUDP mUdp;
	IPAddress mBCast;
	
	int mStatus;	
	unsigned long mLastConnCheck;
	unsigned long mResendTimer;
	unsigned long mHeartbeatTimer;
	unsigned long mHeartbeatTracking[8];
	void (*mConnAchievedCB)();
	void (*mConnLostCB)();
	void (*mMsgReceivedCB)(byte source, unsigned long data);
	void (*mMsgAckdCB)(byte msgId);
	void (*mHeartbeatDiscoveredCB)(byte source);
	void (*mHeartbeatLostCB)(byte source);
	
	
	bool checkConnection();
	void beginConnection();
	void processIncoming();
	void sendMessageHelper(byte recipient, byte msgIdx, unsigned long data);
	void sendMessageAck(byte recipient, byte msgIdx);
	void sendHeartbeat();
	void resendMessages();
};

#endif

